@@ -3,9 +3,7 @@
 # modify it under the same terms as Perl itself.
 
 package Convert::ASN1;
-
-# $Id: ASN1.pm,v 1.29 2003/10/08 14:29:17 gbarr Exp $
-
+$Convert::ASN1::VERSION = '0.33';
 use 5.004;
 use strict;
 use vars qw($VERSION @ISA @EXPORT_OK %EXPORT_TAGS @opParts @opName $AUTOLOAD);
@@ -23,7 +21,6 @@ BEGIN {
   }
 
   @ISA = qw(Exporter);
-  $VERSION = "0.22";
 
   %EXPORT_TAGS = (
     io    => [qw(asn_recv asn_send asn_read asn_write asn_get asn_ready)],
@@ -44,12 +41,13 @@ BEGIN {
   $EXPORT_TAGS{all} = \@EXPORT_OK;
 
   @opParts = qw(
-    cTAG cTYPE cVAR cLOOP cOPT cCHILD cDEFINE
+    cTAG cTYPE cVAR cLOOP cOPT cEXT cCHILD cDEFINE
   );
 
   @opName = qw(
     opUNKNOWN opBOOLEAN opINTEGER opBITSTR opSTRING opNULL opOBJID opREAL
-    opSEQUENCE opSET opUTIME opGTIME opUTF8 opANY opCHOICE opROID opBCD
+    opSEQUENCE opEXPLICIT opSET opUTIME opGTIME opUTF8 opANY opCHOICE opROID opBCD
+    opEXTENSIONS
   );
 
   foreach my $l (\@opParts, \@opName) {
@@ -119,6 +117,15 @@ sub configure {
     Carp::croak("Unsupported encoding format '$opt{encoding}'");
   }
 
+  # IMPLICIT as default for backwards compatibility, even though it's wrong.
+  $self->{options}{tagdefault} = uc($opt{tagdefault} || 'IMPLICIT');
+
+  unless ($self->{options}{tagdefault} =~ /^(?:EXPLICIT|IMPLICIT)$/) {
+    require Carp;
+    Carp::croak("Default tagging must be EXPLICIT/IMPLICIT. Not $opt{tagdefault}");
+  }
+
+
   for my $type (qw(encode decode)) {
     if (exists $opt{$type}) {
       while(my($what,$value) = each %{$opt{$type}}) {
@@ -149,9 +156,9 @@ sub prepare {
   if( ref($asn) eq 'GLOB' ){
     local $/ = undef;
     my $txt = <$asn>;
-    $tree = Convert::ASN1::parser::parse($txt);
+    $tree = Convert::ASN1::parser::parse($txt,$self->{options}{tagdefault});
   } else {
-    $tree = Convert::ASN1::parser::parse($asn);
+    $tree = Convert::ASN1::parser::parse($asn,$self->{options}{tagdefault});
   }
 
   unless ($tree) {
@@ -227,7 +234,7 @@ sub asn_encode_tag {
 	? pack("V",$_[0])
 	: substr(pack("V",$_[0]),0,3)
       : pack("v", $_[0])
-    : chr($_[0]);
+    : pack("C",$_[0]);
 }
 
 
@@ -253,16 +260,27 @@ sub asn_encode_length {
 
 sub decode {
   my $self  = shift;
+  my $ret;
 
   local $SIG{__DIE__};
-  my $ret = eval { 
+  eval {
     my (%stash, $result);
     my $script = $self->{script};
-    my $stash = (1 == @$script && !$self->{script}[0][cVAR]) ? \$result : ($result=\%stash);
+    my $stash = \$result;
+
+    while ($script) {
+      my $child = $script->[0] or last;
+      if (@$script > 1 or defined $child->[cVAR]) {
+        $result = $stash = \%stash;
+        last;
+      }
+      last if $child->[cTYPE] == opCHOICE or $child->[cLOOP];
+      $script = $child->[cCHILD];
+    }
 
     _decode(
 	$self->{options},
-	$script,
+	$self->{script},
 	$stash,
 	0,
 	length $_[0], 
@@ -270,12 +288,10 @@ sub decode {
 	{},
 	$_[0]);
 
-    $result;
-  };
-  if ($@) {
-    $self->{'error'} = $@;
-    return undef;
-  }
+    $ret = $result;
+    1;
+  } or $self->{'error'} = $@ || 'Unknown error';
+
   $ret;
 }
 
@@ -283,7 +299,7 @@ sub decode {
 sub asn_decode_length {
   return unless length $_[0];
 
-  my $len = ord substr($_[0],0,1);
+  my $len = unpack("C",$_[0]);
 
   if($len & 0x80) {
     $len &= 0x7f or return (1,-1);
@@ -299,14 +315,14 @@ sub asn_decode_length {
 sub asn_decode_tag {
   return unless length $_[0];
 
-  my $tag = ord $_[0];
+  my $tag = unpack("C", $_[0]);
   my $n = 1;
 
   if(($tag & 0x1f) == 0x1f) {
     my $b;
     do {
       return if $n >= length $_[0];
-      $b = ord substr($_[0],$n,1);
+      $b = unpack("C",substr($_[0],$n,1));
       $tag |= $b << (8 * $n++);
     } while($b & 0x80);
   }
@@ -317,7 +333,7 @@ sub asn_decode_tag {
 sub asn_decode_tag2 {
   return unless length $_[0];
 
-  my $tag = ord $_[0];
+  my $tag = unpack("C",$_[0]);
   my $num = $tag & 0x1f;
   my $len = 1;
 
@@ -326,7 +342,7 @@ sub asn_decode_tag2 {
     my $b;
     do {
       return if $len >= length $_[0];
-      $b = ord substr($_[0],$len++,1);
+      $b = unpack("C",substr($_[0],$len++,1));
       $num = ($num << 7) + ($b & 0x7f);
     } while($b & 0x80);
   }
@@ -363,9 +379,9 @@ sub i2osp {
     while($num != 0) {
         my $r = $num % $base;
         $num = ($num-$r) / $base;
-        $result .= chr($r);
+        $result .= pack("C",$r);
     }
-    $result ^= chr(255) x length($result) if $neg;
+    $result ^= pack("C",255) x length($result) if $neg;
     return scalar reverse $result;
 }
 
@@ -376,8 +392,8 @@ sub os2ip {
     eval "require $biclass";
     my $base = $biclass->new(256);
     my $result = $biclass->new(0);
-    my $neg = ord($os) >= 0x80
-      and $os ^= chr(255) x length($os);
+    my $neg = unpack("C",$os) >= 0x80
+      and $os ^= pack("C",255) x length($os);
     for (unpack("C*",$os)) {
       $result = ($result * $base) + $_;
     }
